#!/usr/bin/python

#
# Copyright (c) 20202 Waystream AB. All rights reserved.
#
# This Software and its content are protected by the Swedish Copyright Act
# (Sw: Upphovsrattslagen) and, if applicable, the Swedish Patents Act
# (Sw: Patentlagen). All and any copying and distribution of the software
# and its content without authorization by Waystream AB
# is prohibited. The prohibition includes every form of reproduction and
# distribution.
#
#

#
# File:    ibos-cored.py
# By:      Erik Anggard, Waystream AB
# Created: 2014-05-07, Updated 2020-10-27
#
# Daemon for receiving core-files from iBOS.
#
#

from struct import unpack, calcsize
from collections import namedtuple
from socket import *
import getopt, sys
import os, errno
import threading
import select, time
hasgrp = True
try:
    import grp
except:
    hasgrp = False

#
# Defines
#

MAXCOMLENX = 16

CORED_V2_MAGIC = 0xbadc0de2
CORED_V2_VERNAMELEN = 48
MAX_SCRIPT_PATH_LEN = 1024

#struct core_net {
#	uint32_t  version ;    /* iBOS version, old 32bit format */
#	char      processname[MAXCOMLENX] ;
#	/* End of version1 header */
#
#	uint32_t  magic ;      /* Magic (version) */
#	uint32_t  len ;        /* Length of core including this header */
#	char      vername[CORED_V2_VERNAMELEN] ; /* Named iBOS version */
#} __attribute__((packed)) ;

core_net = namedtuple('core_net', 'version processname magic len vername')
core_net.format = "I%dsII%ds" % (MAXCOMLENX, CORED_V2_VERNAMELEN)
core_net.size = calcsize(core_net.format)


#
# Globals
#
dbg = False
gid = -1
script_trigger = None
inc_ip = True
plain_dir = True

#
# main
#
def main():
    global dbg, gid, script_trigger, inc_ip, plain_dir
    p = 10000

    # Handle command line arguments.
    try:
        opts, args = getopt.getopt(sys.argv[1:], "Dg:d:p:s:ic:h")
    except getopt.GetoptError as err:
        print(str(err))
        usage()

    for ch, optarg in opts:
        if ch == '-D':
            dbg = True
        elif ch == '-d':
            try:
                os.chdir(optarg)
            except:
                print("Failed to change to directory '%s'" % optarg)
                sys.exit(1)
            inc_ip = False
            plain_dir = False
        elif ch == '-c':
            try:
                os.chdir(optarg)
            except:
                print("Failed to change to directory '%s'" % optarg)
                sys.exit(1)
        elif ch == '-g':
            if not hasgrp:
                print("Option not available on this system")
                sys.exit(1)                
            try:
                g = grp.getgrnam(optarg)
            except:
                print("Invalid group: %s" % optarg)
                sys.exit(1)
                gid = g[2]
        elif ch == '-p':
            p = int(optarg)
            if p < 1 or p > 0xffff:
                print("Invalid port specified: %s\n" % optarg)
                sys.exit(1)
        elif ch == '-s':
            script_trigger = optarg
        elif ch == '-i':
            inc_ip = 1
        elif ch == '-h' or ch == '-?':
            usage()
        else:
            usage()

    if len(args) > 0:
        usage()

    os.umask(0o007)

    # Setup socket to listen on.
    s = socket(AF_INET, SOCK_STREAM)

    s.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    #s.setsockopt(SOL_SOCKET, SO_REUSEPORT, 1)

    s.bind(("", p))
    s.listen(10)

    if dbg:
        print("ibos-cored started listening on port %d" % p)

    if dbg and gid != -1:
        print("using group id %d" % gid)

    #if not dbg:
    #	daemon(1, 0);

    core_threads = []
    try:
        while True:
            conn, addr = s.accept()
            core_thread = threading.Thread(target=handle_connection, args=[conn, addr])
            core_thread.start()
            core_threads.append(core_thread)
    finally:
        s.close()
        for core_thread in core_threads:
            core_thread.join()


#
# usage
#
def usage():
    s = """
usage: ibos-cored.py [-D] [-d <dir>] [-i] [-g <group>] [-p <port>] [-s <script>]

  ibos-cored will accept connections from ibos elements to receive core files
  and store them on disk.

  -D           Turn on debug mode
  -c <dir>     Store core files in <dir>/<ip>-<file>
  -d <dir>     Store core files in <dir>/<ip>/<file>
  -g <group>   Set group of core file to <group>
  -p <port>    Listen to port <port> (default 10000)
  -s <script>  Run script when a core file has been received (e.g. to send
               notifications)
  -h           Show this help text

file name format: <ip>-<version>.<process>.core.<timestamp>
  <ip>: ip address of unit sending core file
  <version>: ibos version that generated the core file
  <process>: name of process that generated the core file or 'kernel'
  <timestamp>: time the core was received (seconds since Jan 1 1970 in hex)

example:
  python ibos-cored.py -c /var/corefiles
"""
    print(s)
    sys.exit(1)

#
# cstr
#
def cstr(s):
    """ Convert a \0 terminated binary c string to python strig """
    if isinstance(s, str):
        n = s.find('\x00')
        if n != -1:
            return s[:n].decode('utf-8')
        return s.decode('utf-8')
    if isinstance(s, bytes):
        for i in range(0, len(s)):
            if s[i] == 0:
                return s[:i].decode('utf-8')
        return s.decode('utf-8')
    return None

#
# handle_connection
#
def handle_connection(s, addr):
    global dbg, gid, script_trigger, inc_ip, plain_dir
    global core_net
    
    TO = (20*1000)        # 20 seconds timeout
    BUFSZ = (64*1024)
    pfx = ""
    dpfx = ""
    
    dir = str(addr[0])
    if not plain_dir:
        try:
            os.makedirs(dir, 0o700)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(dir):
                pass
            else:
                print("Failed to create directory '%s'" % dir)
                return
        if gid != -1:
            os.chown(dir, -1, gid)
        dpfx = os.path.join(dir,"")

    if inc_ip:
        pfx = "%s-" % dir

    rs, ws, es = select.select([s], [], [s], TO)
    if len(rs) == 0:
        print("Connection timed out from '%s'" % dir)
        return

    data = s.recv(core_net.size)
    if len(data) != core_net.size:
        print("Failed to read core header from '%s'." % dir)
        return

    cn = core_net._make(unpack(core_net.format, data))
    
    if cn.magic != htonl(CORED_V2_MAGIC):
        print("Invalid magic number in core header from '%s'." % dir)
        return

    fn = ".%s%s.%s.core.%x" % (pfx, cstr(cn.vername), cstr(cn.processname),
                               int(time.time()))
    fnp = fn[1:]
    fn = "%s%s" % (dpfx, fn)
    fnp = "%s%s" % (dpfx, fnp)

    f = open(fn, "wb")
    os.chmod(fn, 0o600)
    if gid != -1:
        os.chown(fn, -1, gid)

    if dbg:
        print("Writing file: %s" % fn)

    done = False
    totsz = 0
    while not done:
        rs, ws, es = select.select([s], [], [s], TO)
        if len(rs) == 0:
            print("Connection timed out from '%s'" % dir)
            return
        data = s.recv(BUFSZ)
        if data:
            f.write(data)
            totsz += len(data)
        else:
            # we are done
            done = True

    if dbg:
        print("Closing file: %s" % fn)

    s.close()
    f.close()

    # strip '.' postfix
    if dbg:
        print("Renaming file to: %s " % fnp)
    os.rename(fn, fnp)

    if script_trigger != None:
        cdir = os.getcwd()
        script_exec = "%s %s %s" % (script_trigger, fnp, cdir)
        try:
            os.system(script_exec)
        except:
            print("script exec failed: %s" % script_exec)


#
# call main
#
main()
