/*
 * Copyright (c) 2020 Waystream AB. All rights reserved.
 *
 * This Software and its content are protected by the Swedish Copyright Act
 * (Sw: Upphovsrttslagen) and, if applicable, the Swedish Patents Act
 * (Sw: Patentlagen). All and any copying and distribution of the software
 * and its content without authorization by Waystream AB
 * is prohibited. The prohibition includes every form of reproduction and
 * distribution.
 *
 */

/*
 * File:    ibos-cored.c
 * By:      Erik Anggard, Waystream AB
 * Created: 2001-10-02, Updated 2020-10-27
 *
 * Daemon for receiving core-files from iBOS.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <poll.h>
#include <grp.h>

#ifdef linux
#define strlcpy(a,b,c) do {   \
  strncpy((a), (b), (c) - 1); \
  (a)[(c) - 1] = '\0';        \
} while(0)
#endif

/*
 * Defines
 */

#define MAXCOMLENX  16

#define CORED_V2_MAGIC             0xbadc0de2
#define CORED_V2_VERNAMELEN        48
#define MAX_SCRIPT_PATH_LEN        1024

struct core_net {
	uint32_t  version ;    /* iBOS version, old 32bit format */
	char      processname[MAXCOMLENX] ;
	/* End of version1 header */

	uint32_t  magic ;      /* Magic (version) */
	uint32_t  len ;        /* Length of core including this header */
	char      vername[CORED_V2_VERNAMELEN] ; /* Named iBOS version */
} __attribute__((packed)) ;


/*
 * Prototypes
 */
static void usage(void);
static void handle_connection(int s, struct sockaddr_in si);


/*
 * Globals
 */
int dbg = 0;
gid_t gid = -1;
char *script_trigger = NULL;
int inc_ip = 1;
int plain_dir = 1;

/*
 * main
 */
int
main(int argc, char *argv[])
{
	int s, ns, ch, tmp;
	int p = 10000;
	struct sockaddr_in si, ni;
	pid_t pid;
        struct sigaction nsa;
	socklen_t al;
	extern char *optarg;
	extern int optind;
	struct group *g;

	/* Handle command line arguments. */
	while ((ch = getopt(argc, argv, "Dg:d:p:s:ic:h")) != -1) {
		switch (ch) {
		case 'D':
			dbg = 1;
			break;
		case 'd':
			if (chdir(optarg) == -1) {
				printf("Failed to change to directory "
				       "'%s': %s\n",
				       optarg, strerror(errno));
				exit(1);
			}
			inc_ip = 0;
			plain_dir = 0;
			break;
		case 'c':
			if (chdir(optarg) == -1) {
				printf("Failed to change to directory "
				       "'%s': %s\n",
				       optarg, strerror(errno));
				exit(1);
			}
			break;
		case 'g':
			if (!isdigit(*optarg) || (gid = (gid_t)atoi(optarg)) == -1) {
				if ((g = getgrnam(optarg)) == NULL) {
					printf("Invalid group: %s\n", optarg);
					exit(1);
				}
				gid = g->gr_gid;
			}
			break;
		case 'p':
			p = atoi(optarg);
			if (p < 1 || p > 0xffff) {
				printf("Invalid port specified: %s\n", optarg);
				exit(1);
			}
			break;
		case 's':
			script_trigger = optarg;
			break;
		case 'i':
			inc_ip = 1;
			break;
		case '?':
		case 'h':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (argc)
		usage();

	umask(0007);

	/* Setup socket to listen on. */
	if ((s = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket");
		exit(1);
	}

	tmp = 1;
	setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&tmp, sizeof(tmp));
#ifndef linux
	tmp = 1;
	setsockopt(s, SOL_SOCKET, SO_REUSEPORT, (char *)&tmp, sizeof(tmp));
#endif

	memset(&si, 0, sizeof(si));
#ifndef linux
	si.sin_len = sizeof(si);
#endif
	si.sin_family = PF_INET;
	si.sin_port = htons(p);
	si.sin_addr.s_addr = INADDR_ANY;

	if (bind(s, (struct sockaddr *)&si, sizeof(si))) {
		perror("bind");
		exit(1);
	}
	
	if (listen(s, 10) == -1) {
		perror("listen");
		exit(1);
	}

	if (dbg)
		printf("ibos-cored started listening on port %d\n", p);

	if (dbg && gid != -1)
		printf("using group id %d\n", gid);

	if (!dbg)
		daemon(1, 0);

	/* Setup signal handler for child exit. */
        sigemptyset(&nsa.sa_mask);
        nsa.sa_flags = SA_NOCLDSTOP|SA_NOCLDWAIT;
        nsa.sa_handler = SIG_IGN;
        (void)sigaction(SIGCHLD, &nsa, NULL);

	/* Handle connections. */
	for (;;) {
		al = sizeof(ni);
		if ((ns = accept(s, (struct sockaddr *)&ni, &al)) == -1) {
			perror("accept");
			exit(1);
		}
		tmp = 0;
		setsockopt(ns, SOL_SOCKET, SO_LINGER, (char *)&tmp, 
			   sizeof(tmp));

		if (dbg)
			printf("Got connection from '%s'.\n",
			       inet_ntoa(ni.sin_addr));
		pid = fork();
		if (pid == 0) { /* child */
			handle_connection(ns, ni);
			/* NOTREACHED */
			exit (1);
		} else if (pid == -1) {
			perror("fork");
			exit(-1);
		}
		close(ns);
	}

	close(s);

	return (0);
}


/*
 * usage
 */
static void
usage(void)
{
	fprintf(stderr, 
"usage: ibos-cored [-D] [-d <dir>] [-i] [-g <group>] [-p <port>] [-s <script>]\n"
"\n"
"  ibos-cored will accept connections from ibos elements to receive core files\n"
"  and store them on disk.\n"
"\n"
"  -D           Turn on debug mode\n"
"  -c <dir>     Store core files in <dir>/<ip>-<file>\n"
"  -d <dir>     Store core files in <dir>/<ip>/<file>\n"
"  -g <group>   Set group of core file to <group>\n"
"  -p <port>    Listen to port <port> (default 10000)\n"
"  -s <script>  Run script when a core file has been received (e.g. to send\n"
"               notifications)\n"
"  -h           Show this help text\n"
"\n"
"file name format: <ip>-<version>.<process>.core.<timestamp>\n"
"  <ip>: ip address of unit sending core file\n"
"  <version>: ibos version that generated the core file\n"
"  <process>: name of process that generated the core file or 'kernel'\n"
"  <timestamp>: time the core was received (seconds since Jan 1 1970 in hex)\n"
"\n"
"example:\n"
"  ibos-cored -c /var/corefiles\n"
"\n"
		);

	exit(1);
}


/*
 * handle_connection
 */
static void 
handle_connection(int s, struct sockaddr_in si)
{
#define TO (20*1000)        /* 20 seconds timeout */
#define BUFSZ (64*1024)
	int fd, r, hlen;
	char *dir;
	struct core_net cn;
	char fn[FILENAME_MAX];
	char *fnp = fn;
	char buf[BUFSZ];
	ssize_t sz;
	struct pollfd fds;
	char script_exec[MAX_SCRIPT_PATH_LEN];
	char cdir[MAXPATHLEN] = "";
	char pfx[MAXPATHLEN] = "";
	int ok = 0;
	
	dir = inet_ntoa(si.sin_addr);
	if (!plain_dir) {
		if ((r = mkdir(dir, 0770)) == -1 &&
		    errno != EEXIST) {
			printf("Failed to create directory '%s': %s\n", 
			    dir, strerror(errno));
			exit(1);
		}
		if (r == 0 && gid != -1 )
			chown(dir, (uid_t)-1, gid);

		if (chdir(dir)) {
			printf("Failed to change to directory '%s': %s\n", 
			    dir, strerror(errno));
			exit(1);
		}
	}

	if (inc_ip)
		snprintf(pfx, sizeof(pfx), "%s-", dir);

	fds.fd = s;
	fds.events = POLLIN;
	r = poll(&fds, 1, TO);
	if (r == 0) {
		printf("Connection timed out.\n");
		close(s);
		exit (1);
	} else if (r == -1) {
		perror("poll");
		close(s);
		exit (1);
	}

	if (!(fds.revents & POLLIN) || 
	    ((hlen = read(s, &cn, sizeof(cn))) != sizeof(cn) && 
		hlen != sizeof(cn))) {
		printf("Failed to read core header from '%s'.\n", dir);
		close(s);
		exit(1);		
	}

	if (cn.magic != htonl(CORED_V2_MAGIC)) {
		printf("Invalid magic number in core header from '%s'.\n", dir);
		close(s);
		exit(1);
	}

	snprintf(fn, sizeof(fn), ".%s%s.%s.core.%x",
	    pfx, cn.vername, cn.processname, (int)time(NULL));

	if ((fd = open(fn, O_WRONLY|O_CREAT|O_TRUNC, 0660)) == -1) {
		printf("Failed to create file '%s': %s\n",
		       fn, strerror(errno));
		close(s);
		exit(1);
	}
	if (gid != -1 )
		chown(fn, (uid_t)-1, gid);

	if (dbg)
		printf("Writing file: %s\n", fn);

#ifndef linux
	setproctitle("%s/%s", getcwd(cdir, sizeof(cdir)), fn);
#endif

	do {
		r = poll(&fds, 1, TO);

		if (r == 1 && (fds.revents & POLLIN)) {
			sz = read(s, buf, BUFSZ);
			if (sz == 0) {
				ok = 1;
				break;
			}
			write(fd, buf, sz);
		} else if (r == 0)
			printf("Connection timed out.\n");
		else if (fds.revents & POLLHUP) {
			printf("Connection closed.\n");
			ok = 1;
		} else
			perror("poll");
		
	} while (r == 1);

	if (dbg)
		printf("Closing file: %s\n", fn);
	close(fd);
	close(s);

	if (ok) {
		/* strip '.' postfix */
		fnp++;
		if (dbg)
			printf("Renaming file to: %s\n", fnp);
		if (rename(fn, fnp)) {
			perror("rename");
			exit(1);
		}
	}

	if (ok && script_trigger) {
		getcwd(cdir, sizeof(cdir));
		if (snprintf(script_exec, MAX_SCRIPT_PATH_LEN, "%s %s %s",
			script_trigger, fnp, cdir) > MAX_SCRIPT_PATH_LEN) {
			fprintf(stderr,"script path + image too long "
			    "(max MAX_SCRIPT_PATH_LEN): %s\n", script_exec);
			exit(1);
		}
		if (system(script_exec) == 127) {
			perror("system");
			exit(1);
		} 
	}

	exit(0);
}
